--[[ =================================================================
    Description:
        Type /ubi or /uberinventory to see what you currently own,
        whether it is in the bank or in the bags you carry.
        So now where ever you are in the World of Warcraft you can
        get an overview of all your loot.

        Also vendor sell and buy prices are collected whenever you
        visit a vendor. This means that you can see how much
        something is worth at a vendor. This information is available
        for items in your bags or keyring, your equipped items, items
        within loot frames, action bar items, etc.

    Download:
        UberInventory - http://redeye.xs4all.nl/ubi/

    Contact:
        For questions, bug reports visit the website or send an email
        to the following address: wowaddon@xs4all.nl

    Dependencies:
        None

    Credits:
        A big 'Thank You' to all the people at Blizzard Entertainment
        for making World of Warcraft.
    ================================================================= --]]

-- Text and tooltips for checkbuttons
    UBI_CheckButtons = {};
    UBI_CheckButtons[1] = { ["text"] = UBI_OPTION_MONEY, ["tooltip"] = UBI_OPTION_MONEY_TIP };
    UBI_CheckButtons[2] = { ["text"] = UBI_OPTION_BALANCE, ["tooltip"] = UBI_OPTION_BALANCE_TIP };
    UBI_CheckButtons[3] = { ["text"] = UBI_OPTION_SELLPRICES, ["tooltip"] = UBI_OPTION_SELLPRICES_TIP };
    UBI_CheckButtons[4] = { ["text"] = UBI_OPTION_TAKEMONEY, ["tooltip"] = UBI_OPTION_TAKEMONEY_TIP };
    UBI_CheckButtons[5] = { ["text"] = UBI_OPTION_SHOWMAP, ["tooltip"] = UBI_OPTION_SHOWMAP_TIP };

-- Static dialogs
    StaticPopupDialogs["UBI_CONFIRM_DELETE"] = {
        text = "%s",
        button1 = UBI_YES,
        button2 = UBI_NO,
        OnShow = function( info )
            this.text = info;
        end,
        timeout = 10,
        whileDead = 1,
        exclusive = 1,
        showAlert = 1,
        hideOnEscape = 1
    };

-- Upgrade data structures
    function UberInventory_Upgrade()
        if ( UBI_Options["version"] ~= UBI_VERSION ) then
            -- Show upgrade message
            UberInventory_Message( UBI_UPGRADE );

            -- Upgrade to version 0.9
            if ( UBI_Options["version"] < "0.9" ) then
            end;

            -- Set current version
            UBI_Options["version"] = UBI_VERSION;
        end;
    end;

-- Days since visit
    function UberInventory_DaysSince( visit )
        return math.floor( difftime( time(), visit or time() ) / UBI_SEC_IN_DAY );
    end;

-- Send message to the default chat frame
    function UberInventory_Message( msg, prefix )
        -- Initialize
        local prefixText = "";

        -- Add application prefix
        if ( prefix ) then
            prefixText = C_GREEN.."UBI: "..C_CLOSE;
        end;

        -- Send message to chatframe
        DEFAULT_CHAT_FRAME:AddMessage( prefixText..( msg or "" ) );
    end;

-- Split money (gold, silver, copper)
    function UberInventory_SplitMoney( amount )
        local gold = math.floor( amount / 10000 );
        local silver = math.floor( ( amount - ( gold * 10000 ) ) / 100 );
        local copper = math.floor( ( amount - ( gold * 10000 ) ) - ( silver * 100 ) );

        return gold, silver, copper;
    end;

-- Get money split for cash owned ( gold, silver, copper )
    function UberInventory_GetMoney()
        local gold, silver, copper = UberInventory_SplitMoney( GetMoney() );

        -- Save money
        UBI_Money["Cash"] = GetMoney();

        return gold, silver, copper;
    end;

-- Build moneystring
    function UberInventory_GetMoneyString( money )
        -- Initialize
        local gold, silver, copper = UberInventory_SplitMoney( money );
        local moneystring = "";

        -- Build money string
        if ( gold > 0 ) then moneystring = moneystring.." "..gold.." "..UBI_GOLD_ICON; end;
        if ( silver > 0 ) then moneystring = moneystring.." "..silver.." "..UBI_SILVER_ICON; end;
        if ( copper > 0 or money == 0 ) then moneystring = moneystring.." "..copper.." "..UBI_COPPER_ICON; end;

        return moneystring:trim();
    end;

-- Specialized SetTooltipMoney function (till internal Blizzard code for SetTooltipMoney is fixed)
    function UberInventory_SetTooltipMoney( tooltip, money, type, prefix, suffix )
        -- Initialize
        local moneystring = UberInventory_GetMoneyString( money );

        -- Build the string
        if ( prefix ) then moneystring = prefix..moneystring; end;
        if ( suffix ) then moneystring = moneystring..suffix; end;

        -- Add info to the tooltip
        tooltip:AddLine( moneystring, 1.0, 1.0, 1.0 );
    end;

-- Update slot count information
    function UberInventory_Update_SlotCount()
        UberInventoryFrameBagSlots:SetFormattedText( UBI_SLOT_BAGS, UBI_Options["bag_free"] , UBI_Options["bag_max"] );
        UberInventoryFrameBankSlots:SetFormattedText( UBI_SLOT_BANK, UBI_Options["bank_free"], UBI_Options["bank_max"] );
    end;

-- Send verison information
    function UberInventory_SendVersion()
        -- Send version if in a guild
        if ( IsInGuild() ) then
            SendAddonMessage( "UBI", "UBI:VERSION "..UBI_VERSION, "GUILD" );
        end;
    end;

-- Extract info from tooltip
    function UberInventory_Scan_Tooltip( itemid, color, searchtext )
        -- Rounding function
        function round( x, digit )
            -- Positions to be shifted
            local shift = 10 ^ digit;

            -- Round the number to wanted decimals
            return floor( x * shift + 0.5 ) / shift;
        end;

        -- Initialize
        local txtLeft = "";
        local txtRight = "";
        local digits = 4;
        local colorFound = false;
        local textFound = false;

        -- Setup tooltip for extracting data
        local tooltip = UberInventory_ItemTooltip;
        tooltip:ClearLines();
        tooltip:SetHyperlink( UberInventory_GetLink( itemid ) );

        -- Traverse tooltip lines
        for i = 1, tooltip:NumLines() do
            -- Get text
            local txtLeft = getglobal( tooltip:GetName().."TextLeft"..i ):GetText();
            local txtRight = getglobal( tooltip:GetName().."TextRight"..i ):GetText();

            -- Find the specified color
            if ( color ) then
                -- Check left color
                local r, g, b, a = getglobal( tooltip:GetName().."TextLeft"..i ):GetTextColor();

                -- Check if requested color was used
                if ( ( round( r, digits ) == round( color.r, digits ) ) and
                     ( round( g, digits ) == round( color.g, digits ) ) and
                     ( round( b, digits ) == round( color.b, digits ) ) and
                     txtLeft ) then
                    colorFound = true;
                end;

                -- Check right color
                local r, g, b, a = getglobal( tooltip:GetName().."TextRight"..i ):GetTextColor();

                -- Check if requested color was used
                if ( ( round( r, digits ) == round( color.r, digits ) ) and
                     ( round( g, digits ) == round( color.g, digits ) ) and
                     ( round( b, digits ) == round( color.b, digits ) ) and
                     txtRight ) then
                    colorFound = true;
                end;
            end;

            -- Find the specified text
            if ( searchtext and txtLeft ) then
                textFound = txtLeft:match( searchtext );
            end;
            if ( searchtext and txtRight ) then
                textFound = txtRight:match( searchtext );
            end;

            -- Stop on first blank line or new-line char
            if ( txtLeft == "" or txtLeft:find( "\n" ) ) then
                return ( colorFound or textFound );
            end;
        end;

        return ( colorFound or textFound );
    end;

-- Can item used by current player
    function UberInventory_UsableItem( itemid )
        -- Initialize
        local TOOLTIP_RED = { r = 0.99999779462814, g = 0.12548992037773, b = 0.12548992037773 };

        -- Determine if the color red is used within the tooltip
        local usable = ( not UberInventory_Scan_Tooltip( itemid, TOOLTIP_RED, nil ) ) or
                         UberInventory_Scan_Tooltip( itemid, nil, ITEM_SPELL_KNOWN );

        -- Return result (usable or not)
        return usable;
    end;

-- Build list of usable alts
    function UberInventory_GetAlts()
        -- Initialize
        UBI_Characters = {};

        -- Traverse all players to determine valid alts
        foreach( UBI_Data[UBI_REALM], function( key, value )
            if ( key ~= "Guildbank" ) and
               ( key ~= UBI_PLAYER ) and
               ( ( UBI_Data[UBI_REALM][key]["Options"]["faction"] or "<Unknown>" ) == UBI_FACTION ) then
                -- Add alt to the list
                table.insert( UBI_Characters, key );
            end;
        end );

        -- Sort alts alphabetically
        table.sort( UBI_Characters );
    end;

-- Build list of guildbanks
    function UberInventory_GetGuildbanks()
        -- Initialize
        UBI_Guildbanks = {};

        -- Traverse all guildbanks
        foreach( UBI_Guildbank, function( key, value )
            if ( ( value["Faction"] or "<Unknown>" ) == UBI_FACTION ) then
                -- Add alt to the list
                table.insert( UBI_Guildbanks, key );
            end;
        end );

        -- Sort alts alphabetically
        table.sort( UBI_Guildbanks );
    end;

-- Populate locations dropdown box (Also contains items for alts on the same realm from the same faction (Alliance or Horde))
    function UberInventory_Locations_Init()
        -- Add the basic locations (including Guildbank when member of a guild)
        foreach( UBI_LOCATIONS, function( key, value )
            -- Clear
            UBI_Item = {};

            -- Set new item
            UBI_Item = { ["text"] = value,
                         ["value"] = key,
                         ["icon"] = UBI_LOCATION_TEXTURE[key],
                         ["owner"] = this:GetParent(),
                         ["notCheckable"] = 1,
                         ["func"] = UberInventory_Locations_OnClick };

            -- Add item to dropdown
            UIDropDownMenu_AddButton( UBI_Item );
        end );

        -- Add separator if there are guildbanks to be added
        if ( #UBI_Guildbanks > 0 ) then
            -- Clear
            UBI_Item = {};

            -- Set new item
            UBI_Item = { ["text"] = "",
                         ["isTitle"] = 1,
                         ["owner"] = this:GetParent() };
            UIDropDownMenu_AddButton( UBI_Item );

            UBI_Item = { ["text"] = UBI_TEXT_GUILDBANKS,
                         ["isTitle"] = 1,
                         ["owner"] = this:GetParent(),
                         ["textR"] = 1,
                         ["textG"] = 1,
                         ["textB"] = 0 };
            UIDropDownMenu_AddButton( UBI_Item );
        end;

        -- Add guildbanks (how to determine faction for image?)
        foreach( UBI_Guildbanks, function( key, value )
            -- Clear
            UBI_Item = {};

            -- Set new item
            UBI_Item = { ["text"] = value,
                         ["value"] = key+999,
                         ["icon"] = UBI_LOCATION_TEXTURE[7],
                         ["owner"] = this:GetParent(),
                         ["notCheckable"] = 1,
                         ["func"] = UberInventory_Locations_OnClick };

            -- Add highlight to current players guildbank
            if ( value == UBI_GUILD ) then
                local r, g, b, hex = GetItemQualityColor(5);
                UBI_Item["textR"] = r;
                UBI_Item["textG"] = g;
                UBI_Item["textB"] = b;
            end;

            -- Add item to dropdown (Guildbank only added if in a guild)
            UIDropDownMenu_AddButton( UBI_Item );
        end);

        -- Add separator if there are alts to be added
        if ( #UBI_Characters > 0 ) then
            -- Clear
            UBI_Item = {};

            -- Set new item
            UBI_Item = { ["text"] = "",
                         ["isTitle"] = 1,
                         ["owner"] = this:GetParent() };
            UIDropDownMenu_AddButton( UBI_Item );

            UBI_Item = { ["text"] = UBI_ALT_CHARACTER,
                         ["isTitle"] = 1,
                         ["owner"] = this:GetParent(),
                         ["textR"] = 1,
                         ["textG"] = 1,
                         ["textB"] = 0 };
            UIDropDownMenu_AddButton( UBI_Item );
        end;

        -- Add alts to the list from the same realm and faction (Alliance or Horde)
        foreach( UBI_Characters, function( key, value )
            UBI_Item = { ["text"] = value,
                         ["value"] = key+199,
                         ["icon"] = UBI_LOCATION_TEXTURE[7],
                         ["owner"] = this:GetParent(),
                         ["notCheckable"] = 1,
                         ["func"] = UberInventory_Locations_OnClick };
            UIDropDownMenu_AddButton( UBI_Item );
        end );
   end;

-- Update locations dropdown box
   function UberInventory_Locations_OnClick( )
       UBI_FILTER_LOCATIONS = this.value;
       UBI_PAGE = 0;
       if ( UBI_FILTER_LOCATIONS >= 1000 ) then
           UIDropDownMenu_SetText( UBI_Guildbanks[UBI_FILTER_LOCATIONS - 999], this.owner );
       elseif ( UBI_FILTER_LOCATIONS >= 200 ) then
           UIDropDownMenu_SetText( UBI_Characters[UBI_FILTER_LOCATIONS - 199], this.owner );
       else
           UIDropDownMenu_SetText( UBI_LOCATIONS[UBI_FILTER_LOCATIONS], this.owner );
       end;
       UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
   end;

-- Populate quality dropdown box
    function UberInventory_Quality_Init()
       foreach( UBI_QUALITY, function( key, value )
            -- Clear
            UBI_Item = {};

            -- Set new item
            UBI_Item = { ["text"] = value,
                         ["value"] = key,
                         ["owner"] = this:GetParent(),
                         ["notCheckable"] = 1,
                         ["func"] = UberInventory_Quality_OnClick,
                         ["textR"] = ITEM_QUALITY_COLORS[key-2].r,
                         ["textG"] = ITEM_QUALITY_COLORS[key-2].g,
                         ["textB"] = ITEM_QUALITY_COLORS[key-2].b };

            -- Correct color for 'All qualities'
            if (key-2 == -1) then
                UBI_Item["textR"] = ITEM_QUALITY_COLORS[1].r;
                UBI_Item["textG"] = ITEM_QUALITY_COLORS[1].g;
                UBI_Item["textB"] = ITEM_QUALITY_COLORS[1].b;
            end;

            -- Add item to dropdown
            UIDropDownMenu_AddButton( UBI_Item );
        end );
    end;

-- Update quality dropdown box
   function UberInventory_Quality_OnClick( )
       UBI_FILTER_QUALITY = this.value;
       UBI_PAGE = 0;
       UIDropDownMenu_SetText( UBI_QUALITY[UBI_FILTER_QUALITY], this.owner );
       UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
   end;

-- Populate classes dropdown box
    function UberInventory_Classes_Init()
        foreach( UBI_CLASSES, function( key, value )
            -- Clear
            UBI_Item = {};

            -- Set new item
            UBI_Item = { ["text"] = value,
                         ["value"] = key,
                         ["owner"] = this:GetParent(),
                         ["notCheckable"] = 1,
                         ["func"] = UberInventory_Classes_OnClick };

            -- Add item to dropdown
            UIDropDownMenu_AddButton( UBI_Item );
        end );
    end;

-- Update classes dropdown box
   function UberInventory_Classes_OnClick( )
       UBI_FILTER_CLASSES = this.value;
       UBI_PAGE = 0;
       UIDropDownMenu_SetText( UBI_CLASSES[UBI_FILTER_CLASSES], this.owner );
       UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
   end;

-- OnLoad scroll frame
    function UberInventory_Scroll_OnLoad( scrollbar )
        UBI_SCROLLBAR = getglobal( scrollbar );
        UBI_SCROLLBAR:SetValueStep( 1 );
        UBI_SCROLLBAR.scrollBarHideable = true;
    end;

-- Toggle scroll buttons
    function UberInventory_ToggleButtons( pages )
        -- Enable/disable scroll buttons
        local min, max = UBI_SCROLLBAR:GetMinMaxValues();
        if ( min == UBI_SCROLLBAR:GetValue() ) then
            getglobal( UBI_SCROLLBAR:GetName().."ScrollUpButton" ):Disable();
        else
            getglobal( UBI_SCROLLBAR:GetName().."ScrollUpButton" ):Enable();
        end;
        if ( max == UBI_SCROLLBAR:GetValue() or pages < 0 ) then
            getglobal( UBI_SCROLLBAR:GetName().."ScrollDownButton" ):Disable();
        else
            getglobal( UBI_SCROLLBAR:GetName().."ScrollDownButton" ):Enable();
        end;
    end;

-- Initialize scroll frame
    function UberInventory_Scroll_Init( pages )
        -- Initiliaze some settings
        if ( pages or 0 ) < 0 then pages = 0 end;

        -- Set new scrollbar values
        UBI_SCROLLBAR:SetMinMaxValues( 0, pages * UBI_PIX_PER_PAGE );
        UBI_SCROLLBAR:SetValue( UBI_PAGE * UBI_PIX_PER_PAGE );

        -- Update scrollbuttons
        UberInventory_ToggleButtons( pages );
    end;

-- Handle click events from scroll frame
    function UberInventory_Scroll_Click()
        -- Calculate current page
        UBI_PAGE = math.ceil( UBI_SCROLLBAR:GetValue() / UBI_PIX_PER_PAGE );

        -- Update inventory frame
        UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

        -- Update scrollbuttons
        UberInventory_ToggleButtons( UBI_PAGE );
    end;

-- Returns an itemTexture
    function UberInventory_GetTexture( itemId )
        -- Get texture of item
        local itemTexture = GetItemIcon( itemId );

        -- Return texture
        return itemTexture;
    end;

-- Returns an itemLink
    function UberInventory_GetLink( itemId )
        -- Get name of item
        local itemName, _, itemQuality = GetItemInfo( itemId );

        -- If the item has never been seen before, itemName will be nil
        if ( itemName == nil ) then
            itemName = UBI_ITEM_UNKNOWN;
            itemQuality = 0;
        end;

        -- Determine item color
        local color = ITEM_QUALITY_COLORS[itemQuality].hex;

        -- Return itemLink
        return color.."|Hitem:"..itemId..":0:0:0:0:0:0:0|h["..itemName.."]|h|r";
    end;

-- Handle OnEnter event of the Minimap icon
    function UberInventory_Minimap_OnEnter()
        -- Set location of tooltip
        GameTooltip:SetOwner( this, "ANCHOR_BOTTOMLEFT" );

        -- Add title (AddOn name and version)
        GameTooltip:AddDoubleLine( UBI_NAME, UBI_VERSION,
                                   HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b,
                                   NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b );

        -- Add slot information
        GameTooltip:AddDoubleLine( UBI_BAG, UBI_FREE:format( UBI_Options["bag_free"] , UBI_Options["bag_max"] ) );
        GameTooltip:AddDoubleLine( UBI_BANK, UBI_FREE:format( UBI_Options["bank_free"] , UBI_Options["bank_max"] ) );

        -- Add cash inforamtion
        GameTooltip:AddDoubleLine( UBI_CASH_TOTAL, UberInventory_GetMoneyString( GetMoney() + UBI_Money["mail"] ) );

        -- (Re)display the tooltip
        GameTooltip:Show();
    end;

-- Sorting table content
    function UberInventory_SortFilter( str, cmdline )
        -- Initialize
        local include_item = true;

        -- Clear the array and inventory count
        UBI_Sorted = {};
        UBI_Inventory_count = 0;
        UBI_Items_Work = {};

        -- Set base list of items to use
        if ( cmdline or UBI_FILTER_LOCATIONS <= 6 ) then
            UBI_Items_Work = UBI_Items;
            UBI_SHOWING_GUILDBANK = false;
        elseif ( UBI_FILTER_LOCATIONS >= 1000 ) then
            UBI_Items_Work = UBI_Guildbank[UBI_Guildbanks[UBI_FILTER_LOCATIONS - 999]]["Items"];
            UBI_SHOWING_GUILDBANK = true;
        elseif ( UBI_FILTER_LOCATIONS >= 200 ) then
            UBI_Items_Work = UBI_Data[UBI_REALM][UBI_Characters[UBI_FILTER_LOCATIONS - 199]]["Items"];
            UBI_SHOWING_GUILDBANK = false;
        end;

        -- Copy items to local array (including filter)
        foreach( UBI_Items_Work, function ( key, record )
            -- Initialize
            include_item = true;

            -- Count number of unique items
            UBI_Inventory_count = UBI_Inventory_count + 1;

            -- Fix totalCount (only for actual characters)
            if ( record.bag_count ) then
                record.total = record.bag_count + record.bank_count + record.keyring_count +
                               record.mailbox_count + record.equip_count;
            end;

            -- Only apply filters when searching from then inventory frame
            if ( not cmdline ) then
                -- Filter on location (1 = All, 2 = Bags, 3 = Banks, 4 = Keyring, 5 = Mail, 6 = Equipped )
                if ( ( UBI_FILTER_LOCATIONS == 2 and record.bag_count == 0 ) or
                     ( UBI_FILTER_LOCATIONS == 3 and record.bank_count == 0 ) or
                     ( UBI_FILTER_LOCATIONS == 4 and record.keyring_count == 0 ) or
                     ( UBI_FILTER_LOCATIONS == 5 and record.mailbox_count == 0 ) or
                     ( UBI_FILTER_LOCATIONS == 6 and record.equip_count == 0 ) ) then
                    include_item = false;
                end;

                -- Filter on quality
                if ( UBI_FILTER_QUALITY > 1 and record.quality ~= UBI_FILTER_QUALITY-2 ) then
                    include_item = false;
                end;

                -- Filter on class
                if ( UBI_FILTER_CLASSES > 1 and record.type ~= UBI_CLASSES[UBI_FILTER_CLASSES] ) then
                    include_item = false;
                end;

                -- Filter on usability
                if ( UBI_FILTER_USABLE ) then
                    if ( not UberInventory_UsableItem( record.itemid ) ) then
                        include_item = false;
                    end;
                end;
            end;

            -- Filter on name
            if ( ( string.find( string.lower( record.name ) , string.lower( str ) ) or 0 ) > 0 ) and include_item then
                -- Add item to the sorted structure
                table.insert( UBI_Sorted, record );
            end;

        end );

        -- Sort the array alphabetically based on item name
        table.sort( UBI_Sorted, function( rec1, rec2 )
                                    return rec1.name < rec2.name
                                end  );
    end;

-- Reset filters
    function UberInventory_ResetFilters()
        -- Reset current page
        UBI_PAGE = 0;

        -- Reset dropdown for location
        UBI_FILTER_LOCATIONS = 1;
        UIDropDownMenu_SetText( UBI_LOCATIONS[UBI_FILTER_LOCATIONS], getglobal( "UberInventoryFrameLocationDropDown" ) );

        -- Reset dropdown for quality
        UBI_FILTER_QUALITY = 1;
        UIDropDownMenu_SetText( UBI_QUALITY[UBI_FILTER_QUALITY], getglobal( "UberInventoryFrameQualityDropDown" ) );

        -- Reset dropdown for class
        UBI_FILTER_CLASSES = 1;
        UIDropDownMenu_SetText( UBI_CLASSES[UBI_FILTER_CLASSES], getglobal( "UberInventoryFrameClassDropDown" ) );

        -- Reset search text (and refresh items displayed)
        if ( UBI_FILTER_TEXT == "" ) then
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
        else
            UberInventoryFrameItemSearch:SetText("");
        end;
    end;

-- Reset item counts
    function UberInventory_ResetCount( location )
        for key, record in pairs( UBI_Items ) do
            if ( location == "bag" ) then
                record.bag_count = 0;
            elseif ( location == "bank" ) then
                record.bank_count = 0;
            elseif ( location == "keyring" ) then
                record.keyring_count = 0;
            elseif ( location == "mailbox" ) then
                record.mailbox_count = 0;
            elseif ( location == "equipped" ) then
                record.equip_count = 0;
            end;
        end;
    end;

-- Update merchant buy prices
    function UberInventory_MerchantBuyPrices()
        for index = 1, GetMerchantNumItems() do
            local itemName, _, itemPrice, itemQuantity, _, _, _ = GetMerchantItemInfo( index );
            local itemLink = GetMerchantItemLink( index );
            if ( itemLink ) then
                local _, itemId, _, _, _, _, _, _, _ = strsplit( ":", itemLink );

                -- Add item if not already available
                if ( not UBI_Prices[itemId] ) then
                    UBI_Prices[itemId] = { };
                end;

                -- Add buy price to the list
                UBI_Prices[itemId].buy = itemPrice / itemQuantity;
            end;
        end;
    end;

-- Collect merchant sell prices
    function UberInventory_MerchantSellPrices()
        -- Initialize
        local containers = {};

        -- Remove old information
        UberInventory_ItemTooltip:ClearLines();

        -- Bags and Keyring
        if ( HasKey() == 1 ) then
            containers = { KEYRING_CONTAINER, 0, 1, 2, 3, 4 };
        else
            containers = { 0, 1, 2, 3, 4 };
        end;

        -- Traverse bags
        for key, containerID in pairs( containers ) do
            if ( GetContainerNumSlots( containerID ) ) then
                for slotID = 1, GetContainerNumSlots( containerID ) do
                    -- Get itemLink
                    local itemLink = GetContainerItemLink( containerID, slotID );

                    if ( itemLink ) then
                        -- Get current item count
                        _, currentItemCount, _, _, _ = GetContainerItemInfo( containerID, slotID );

                        -- Get current itemId
                        _, currentItemId, _, _, _, _, _, _, _ = strsplit( ":", itemLink );

                        -- Get the price
                        if ( containerID == KEYRING_CONTAINER ) then
                            local hasItem, hasCooldown, repairCost = UberInventory_ItemTooltip:SetInventoryItem( "player", KeyRingButtonIDToInvSlotID( slotID ) );
                        else
                            local hasCooldown, repairCost = UberInventory_ItemTooltip:SetBagItem( containerID, slotID );
                        end;
                    end;
                end;
            end;
        end;
    end;

-- Update Merchant buy/sell prices
    function UberInventory_MerchantUpdate()
        -- Update merchant buy/sell prices
        UberInventory_MerchantBuyPrices();

        -- Update merchant sell prices
        UberInventory_MerchantSellPrices();
        return;
    end;

-- HOOK - New handler for the GameTooltip:OnTooltipAddMoney (fired when at a merchant)
    function UberInventory_OnTooltipAddMoney( self, cost )
        -- Call default function
        local returnValue = UberInventory_ItemTooltip_Original( self, cost );

        -- Store sell price
        if ( currentItemId and currentItemCount ) then
            -- Add item if not already available
            if ( not UBI_Prices[currentItemId] ) then
                UBI_Prices[currentItemId] = { };
            end;

            -- Add buy price to the list
            UBI_Prices[currentItemId].sell = arg1 / currentItemCount;
        end;

        return returnValue;
    end;

-- Add sell prices to items in your bag (only used if not already at a merchant)
    function UberInventory_AddSellPrice( tooltip )
        -- Function for adding counts per location (only non-zero locations are added to the tooltip)
        function AddLocationInfo( tooltip, bag, bank, keyring, mailbox, equipped )
            tooltip:AddLine( "|n" );
            if ( bag > 0 ) then tooltip:AddLine( UBI_LOCATIONS[2].." = "..bag ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[2] ); end;
            if ( bank > 0 ) then tooltip:AddLine( UBI_LOCATIONS[3].." = "..bank ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[3] ); end;
            if ( keyring > 0 ) then tooltip:AddLine( UBI_LOCATIONS[4].." = "..keyring ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[4] ); end;
            if ( mailbox > 0 ) then tooltip:AddLine( UBI_LOCATIONS[5].." = "..mailbox ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[5] ); end;
            if ( equipped > 0 ) then tooltip:AddLine( UBI_LOCATIONS[6].." = "..equipped ); tooltip:AddTexture( UBI_LOCATION_TEXTURE[6] ); end;
        end;

        -- Initialize
        local itemPrice = 0;
        local itemCount = 1;
        local guildCount = 0;
        local spacerAdded = false;

        -- Get the itemLink
        local itemName, itemLink = tooltip:GetItem();

        -- Does the slot contain an item
        if ( itemLink ) then
            -- Get item ID
            local _, itemId, _, _, _, _, _, _, _ = strsplit( ":", itemLink );
            local _, _, _, _, _, itemType, itemSubType = GetItemInfo( itemLink );

            -- Get count (only if count label exists and is visible, otherwise use the value 1)
            if ( tooltip:GetName() == "ItemRefTooltip" ) then
                itemCount = 1;
            else
                -- Special fix needed for Auctioneer Advanced, since this AddOn removes itembutton as soon as
                -- the mouse is over it to display a bigger button on a different location
                if ( GetMouseFocus():GetName() ) then
                  local objCount = getglobal( GetMouseFocus():GetName().."Count" );
                  if ( objCount and objCount:IsVisible() ) then
                      itemCount = tonumber( objCount:GetText() ) or 1;
                  end;
                else
                    itemCount = 1;
                end;
            end;

            -- Get stored sell price
            if ( UBI_Prices[itemId] ) then
               itemPrice = UBI_Prices[itemId]["sell"];
            end;

            -- Only perform the following part when it not has already been done (gets called twice for recipe related items)
            if ( not tooltip.UBI_InfoAdded or false ) then
                -- Add information to the tooltip (not at Merchant and not when option 'Show sell prices' is turned off)
                if ( not MerchantFrame:IsVisible() and UBI_Options["show_sell_prices"] ) or ( tooltip == ItemRefTooltip ) then
                    if ( ( itemPrice or 0 ) == 0 ) then  -- Item unsellable
                        tooltip:AddLine( "|n" );
                        tooltip:AddLine( ITEM_UNSELLABLE, 1.0, 1.0, 1.0 );
                    elseif ( itemCount > 0 ) then -- Calculate sell price and add it
                        -- Add money to the tooltip
                        tooltip:AddLine( "|n" );
                        UberInventory_SetTooltipMoney( tooltip, itemPrice * itemCount, "STATIC", UBI_ITEM_SELL, "" );
                    end;
                end;

                -- Add counts (if UBI_TooltipItem is populated then we are showing a tooltip from the main inventory frame)
                if ( UBI_TooltipItem ) then
                    -- Only add the info when we are not show data from the guildbank
                    if ( UBI_TooltipItem["bag_count"] ) then
                        -- Add full item count information
                        AddLocationInfo( tooltip,
                                         UBI_TooltipItem["bag_count"],
                                         UBI_TooltipItem["bank_count"],
                                         UBI_TooltipItem["keyring_count"],
                                         UBI_TooltipItem["mailbox_count"],
                                         UBI_TooltipItem["equip_count"] );
                        spacerAdded = true;
                    end;
                elseif ( UBI_Items[itemId] ) then
                    AddLocationInfo( tooltip,
                                     UBI_Items[itemId]["bag_count"],
                                     UBI_Items[itemId]["bank_count"],
                                     UBI_Items[itemId]["keyring_count"],
                                     UBI_Items[itemId]["mailbox_count"],
                                     UBI_Items[itemId]["equip_count"] );
                    spacerAdded = true;
                end;

                -- Add info from current character if viewing data from an alt
                if ( UBI_Items[itemId] and UBI_TooltipLocation and UBI_FILTER_LOCATIONS > 6 ) then
                    -- Insert spacer if not already added
                    if ( not spacerAdded ) then
                        tooltip:AddLine( "|n" );
                        spacerAdded = true;
                    end;

                    -- Add the count and icon for the current alt
                    tooltip:AddLine( UBI_PLAYER.." = "..UBI_Items[itemId].total );
                    tooltip:AddTexture( UBI_LOCATION_TEXTURE[7] );
                end;

                -- Add information for alt characters
                foreach( UBI_Characters, function( key, value )
                    if ( UBI_Data[UBI_REALM][value]["Items"][itemId] ) then
                        if ( ( key + 199 ) ~= UBI_TooltipLocation ) then
                            -- Insert spacer if not already added
                            if ( not spacerAdded ) then
                                tooltip:AddLine( "|n" );
                                spacerAdded = true;
                            end;

                            -- Fix totalCount
                            local record = UBI_Data[UBI_REALM][value]["Items"][itemId];
                            record.total = record.bag_count + record.bank_count + record.keyring_count +
                                           record.mailbox_count + record.equip_count;

                            -- Add the count and icon for the current alt
                            tooltip:AddLine( value.." = "..record.total );
                            tooltip:AddTexture( UBI_LOCATION_TEXTURE[7] );
                        end;
                    end;
                end );

                -- Add information for guildbanks
                foreach( UBI_Guildbanks, function( key, value )
                    if ( UBI_Guildbank[value]["Items"][itemId] ) then
                        -- Insert spacer if not already added
                        if ( not spacerAdded ) then
                            tooltip:AddLine( "|n" );
                            spacerAdded = true;
                        end;

                        -- Add the count and icon for the current guildbank
                        tooltip:AddLine( C_GREY..value.." = "..UBI_Guildbank[value]["Items"][itemId].total..C_CLOSE );
                        tooltip:AddTexture( UBI_LOCATION_TEXTURE[7] );
                    end;
                end );

            end;
        end;
    end;

-- HOOK - Set new OnTooltipSetItem and OnHide script for Tooltip objects
    function UberInventory_HookTooltip( tooltip )
        -- Store default script
        local tooltipName = tooltip:GetName();
        UBI_Hooks["OnTooltipSetItem"][tooltipName] = tooltip:GetScript( "OnTooltipSetItem" );
        UBI_Hooks["OnTooltipCleared"][tooltipName] = tooltip:GetScript( "OnTooltipCleared" );

        -- Set new script to handle OntooltipSetItem
        tooltip:SetScript( "OnTooltipSetItem", function( self, ... )
            -- Get tooltip name
            local tooltipName = self:GetName();

            -- Call default script
            if ( UBI_Hooks["OnTooltipSetItem"][tooltipName] ) then
                UBI_Hooks["OnTooltipSetItem"][tooltipName]( self, ... );
            end;

            -- Call new script (adds the price information)
            UberInventory_AddSellPrice( self );

            -- Turn on UberInventory indicator
            self.UBI_InfoAdded = true;
        end );

        -- Set new script to handle OnTooltipCleared
        tooltip:SetScript( "OnTooltipCleared", function( self, ... )
            -- Get tooltip name
            local tooltipName = self:GetName();

            -- Call default script
            if ( UBI_Hooks["OnTooltipCleared"][tooltipName] ) then
                UBI_Hooks["OnTooltipCleared"][tooltipName]( self, ... );
            end;

            -- Turn off UberInventory indicator
            self.UBI_InfoAdded = false;
        end );
     end;

-- Get item information and store it
    function UberInventory_Item( bagID, slotID, location )
        -- Initialize
        local bankCount = 0;
        local bagCount = 0;
        local keyCount = 0;
        local mailCount = 0;
        local equipCount = 0;
        local itemBuyPrice = 0;
        local itemSellPrice = 0;
        local itemLink = nil;

        -- Get itemLink
        if ( location == "mailbox" ) then
            itemLink = GetInboxItemLink( bagID, slotID );
        elseif ( location == "equipped" ) then
            itemLink = GetInventoryItemLink( "player", slotID );
        else
            itemLink = GetContainerItemLink( bagID, slotID );
        end;

        -- Retrieve item information
        if ( itemLink ) then
            -- Initialize
            local itemCount = 0;

            -- Fetch item information (itemLink skipped from GetItemInfo, does not always work at this stage (Quest related items))
            local itemName, _, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo( itemLink );
            local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit( ":", itemLink );

            if ( location == "mailbox" ) then
                _, _, itemCount = GetInboxItem( bagID, slotID );
            elseif ( location == "equipped" ) then
                itemCount = 1;
            else
                _, itemCount = GetContainerItemInfo( bagID, slotID );
            end;

            -- Get existing item counts or activate sorting (new item is added)
            if ( UBI_Items[itemId] ) then
                bankCount = UBI_Items[itemId]["bank_count"];
                bagCount = UBI_Items[itemId]["bag_count"];
                keyCount = UBI_Items[itemId]["keyring_count"];
                mailCount = UBI_Items[itemId]["mailbox_count"];
                equipCount = UBI_Items[itemId]["equip_count"];
            end;

            -- Sort out itemCount
            if ( location == "bank" ) then
                bankCount = bankCount + itemCount;
            elseif ( location == "bag" ) then
                bagCount = bagCount + itemCount;
            elseif ( location == "keyring" ) then
                keyCount = keyCount + itemCount;
            elseif ( location == "mailbox" ) then
                mailCount = mailCount + itemCount;
            elseif ( location == "equipped" ) then
                equipCount = equipCount + itemCount;
            end;
            local totalCount = bagCount + bankCount + keyCount + mailCount + equipCount;

            -- Store the item information
            UBI_Items[itemId] = { ["itemid"] = itemId,
                                  ["name"] = itemName,
                                  ["quality"] = itemQuality,
                                  ["level"] = itemLevel,
                                  ["bag_count"] = bagCount,
                                  ["bank_count"] = bankCount,
                                  ["keyring_count"] = keyCount,
                                  ["mailbox_count"] = mailCount,
                                  ["equip_count"] = equipCount,
                                  ["type"] = itemType,
                                  ["subtype"] = itemSubType,
                                  ["total"] = totalCount };
        end;
    end;

-- Handle Bag items
    function UberInventory_Save_Bag()
        -- Initialize
        local slotCount, freeCount = 0, 0;

        -- Reset bag counts
        UberInventory_ResetCount( "bag" );

        -- Traverse bags
        for bagID = 0, 4 do
            if ( GetContainerNumSlots( bagID ) ) then
                slotCount = slotCount + GetContainerNumSlots( bagID );
                freeCount = freeCount + GetContainerNumFreeSlots( bagID );
                for slotID = 1, GetContainerNumSlots( bagID ) do
                    UberInventory_Item( bagID, slotID, "bag" );
                end;
            end;
        end;

        -- Save slot count
        UBI_Options["bag_max"] = slotCount;
        UBI_Options["bag_free"] = freeCount;
    end;

-- Handle Bank items
    function UberInventory_Save_Bank()
        -- Intialize
        local slotCount, freeCount = 0, 0;

        if ( UBI_BANK_OPEN ) then
            -- Reset bank counts
            UberInventory_ResetCount( "bank" );

            -- Traverse bank slots
            slotCount = GetContainerNumSlots( BANK_CONTAINER );
            for slotID = 1, GetContainerNumSlots( BANK_CONTAINER ) do
                if ( not GetContainerItemLink( BANK_CONTAINER, slotID ) ) then
                    freeCount = freeCount + 1;
                end;
                UberInventory_Item( BANK_CONTAINER, slotID, "bank" );
            end;

            -- Traverse bank bags
            for bagID = 5, 11 do
                if ( GetContainerNumSlots( bagID ) ) then
                    slotCount = slotCount + GetContainerNumSlots( bagID );
                    for slotID = 1, GetContainerNumSlots( bagID ) do
                        if ( not GetContainerItemLink( bagID, slotID ) ) then
                            freeCount = freeCount + 1;
                        end;
                        UberInventory_Item( bagID, slotID, "bank" );
                    end;
                end;
            end;

            -- Save slot count
            UBI_Options["bank_max"] = slotCount;
            UBI_Options["bank_free"] = freeCount;
        end;
    end;

-- Handle Keyring items
    function UberInventory_Save_Keyring()
        -- Reset keyring counts
        UberInventory_ResetCount( "keyring" );

        -- Traverse keyring
        if ( HasKey() == 1 ) then
            for slotID = 1, GetContainerNumSlots( KEYRING_CONTAINER ) do
                UberInventory_Item( KEYRING_CONTAINER, slotID, "keyring" );
            end;
        end;
    end;

-- Handle Mailbox items
    function UberInventory_Save_Mailbox()
        -- Initialize
        local mailCash = 0;

        if ( UBI_MAILBOX_OPEN ) then
            -- Reset mailbox counts
            UberInventory_ResetCount( "mailbox" );

            -- Exit if inbox is empty (also set mailbox cash to zero)
            if ( GetInboxNumItems() == 0 ) then
               UBI_Money["mail"] = mailCash;
               return;
            end;

            -- Traverse mailbox (Reverse order)
            for mailID = GetInboxNumItems(), 1, -1 do
                -- Mail info
                local _, _, mailSender, mailSubject, cashAttached, CODAmount, daysLeft, itemAttached, mailRead, mailReturned, textCreated, canReply, isGM = GetInboxHeaderInfo( mailID );

                -- Handle attachment
                if ( itemAttached ) then
                    for attachID = 1, 12 do
                        UberInventory_Item( mailID, attachID, "mailbox" );
                    end;
                end;

                -- Handle money
                if ( cashAttached > 0 ) then
                    if ( UBI_Options["take_money"] ) then
                       -- Get mail text to mark the messages as read
                       local mailText, mailTexture, mailTakeable, mailInvoice = GetInboxText( mailID );

                        if ( cashAttached > 0 ) then
                            -- Take the money, if not items are attached this also deletes the mail
                            TakeInboxMoney( mailID );

                            -- Report cash collected (turned off for now, issue with duplicate events being fired)
                            --UberInventory_Message( (event or "").." >> "..UBI_MAIL_CASH:format( mailSender, mailSubject, UberInventory_GetMoneyString( cashAttached ) ) );

                        end;
                    else
                        -- Record the mailbox cash
                        mailCash = mailCash + cashAttached;
                    end;
                end;
            end;

            -- Save mail cash
            UBI_Money["mail"] = mailCash;

            -- Update wallet/mail info (if mainframe is shown)
            UberInventory_WalletMailCashInfo();
        end;
    end;

-- Handle Equiped items
    function UberInventory_Save_Equipped()
        -- Reset Equiped counts
        UberInventory_ResetCount( "equipped" );

        -- Traverse equip slots
        foreach( UBI_EQUIP_SLOTS, function( key, value )
            local slotID, _ = GetInventorySlotInfo( value );
            UberInventory_Item( nil, slotID, "equipped" );
        end );

    end;

-- Handle Guildbank items
    function UberInventory_Save_Guildbank( event )
        -- Initialize
        local itemCount = 0;

        if ( UBI_GUILDBANK_OPEN ) then
            -- Clear the existing data
            UBI_Guildbank[UBI_GUILD]["Items"] = {};

            -- Traverse all guildbank tabs and slots
            for gbTab = 1, GetNumGuildBankTabs() do
                -- Retrieve data from the server
                if ( event ~= "GUILDBANKBAGSLOTS_CHANGED" ) then
                    QueryGuildBankTab( gbTab );
                end;

                for gbSlot = 1, MAX_GUILDBANK_SLOTS_PER_TAB do
                    -- Reset
                    itemCount = 0;

                    -- Get slot information
                    local _, count, _ = GetGuildBankItemInfo( gbTab, gbSlot );
                    local itemLink = GetGuildBankItemLink( gbTab, gbSlot );

                    if ( itemLink ) then
                        -- Fetch item information
                        local itemName, _, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo( itemLink );
                        local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit( ":", itemLink );

                        -- Get existing item count
                        if ( UBI_Guildbank[UBI_GUILD]["Items"][itemId] ) then
                            itemCount = UBI_Guildbank[UBI_GUILD]["Items"][itemId]["count"];
                        end;

                        -- Store the item information
                        UBI_Guildbank[UBI_GUILD]["Items"][itemId] = { ["itemid"] = itemId,
                                                                      ["name"] = itemName,
                                                                      ["quality"] = itemQuality,
                                                                      ["level"] = itemLevel,
                                                                      ["count"] = itemCount + count,
                                                                      ["type"] = itemType,
                                                                      ["subtype"] = itemSubType,
                                                                      ["total"] = itemCount + count };
                    end;
                end;
            end;
        end;

        -- Rebuild list of guildbanks
        UberInventory_GetGuildbanks();

        -- Update inventory frame if visible
        if ( UberInventoryFrame:IsVisible() ) then
            -- Update items
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;

        -- Perform garbage collection
        collectgarbage( "collect" );
    end;

-- Handle Guildbank items
    function UberInventory_Receive_Guildbank()
        -- Init guildbank structures (if needed)
        UberInventory_Guildbank_Init()

        -- Show receiving data
        UberInventory_Message( UBI_NEW_GBDATA:format( UBI_GBSender ) );

        -- Clear the existing data
        UBI_Guildbank[UBI_GUILD]["Items"] = {};

        -- Traverse all guildbank tabs and slots
        foreach( UBI_GBData, function ( key, value )
            -- Fetch item information
            local itemLink = UberInventory_GetLink( value.itemid );
            local itemName, _, itemQuality, itemLevel, itemMinLevel, itemType, itemSubType, itemStackCount, itemEquipLoc, itemTexture = GetItemInfo( itemLink );

            -- Store the item information
            UBI_Guildbank[UBI_GUILD]["Items"][value.itemid] = { ["itemid"] = value.itemid,
                                                                ["name"] = value.name or itemName or UBI_UNCACHED,
                                                                ["quality"] = value.quality or itemQuality,
                                                                ["level"] = value.level or itemLevel or 0,
                                                                ["count"] = value.count,
                                                                ["type"] = value.type or itemType or UBI_UNCACHED,
                                                                ["subtype"] = value.subtype or itemSubType or UBI_UNCACHED,
                                                                ["total"] = value.count };
            end );

        -- Update inventory frame if visible
        if ( UberInventoryFrame:IsVisible() ) then
            -- Update items
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;

        -- Track Guildbank last visit date/time
        UBI_Options["guildbank_visit"] = time();

        -- Perform garbage collection
        collectgarbage( "collect" );

        -- Clear sender
        UBI_GBSender = nil;
    end;

-- Send Guildbank data to other guild members
    function UberInventory_Send_Guildbank()
        -- Initialize
        local sendData = "";

        -- Send start message
        SendAddonMessage( "UBI", "UBI:GBSTART "..UBI_VERSION, "GUILD" );

        -- Send the actual guildbank data
        foreach( UBI_Guildbank[UBI_GUILD]["Items"], function ( key, record )
            -- Create data string
            sendData = key.." "..
                       record["count"].." "..
                       string.gsub( record["name"], " ", "_" ).." "..
                       string.gsub( record["type"], " ", "_" ).." "..
                       string.gsub( record["subtype"], " ", "_" ).." "..
                       record["level"].." "..
                       record["quality"];

            -- Send data string
            SendAddonMessage( "UBI", "UBI:GBITEM "..sendData, "GUILD" );
        end );

        -- Send cash
        SendAddonMessage( "UBI", "UBI:GBCASH "..UBI_Guildbank[UBI_GUILD]["Cash"], "GUILD" );

        -- Send done message
        SendAddonMessage( "UBI", "UBI:GBEND", "GUILD" );
    end;

-- Cleanup inventory
    function UberInventory_Cleanup()
        -- Remove items for which bag_count, bank_count, keyring_count, mailbox_count and equip_count equals 0
        for key, record in pairs( UBI_Items ) do
            if ( record.name ) then
                if ( record.bag_count + record.bank_count + record.keyring_count + record.mailbox_count + record.equip_count == 0 ) then
                    -- Remove the item
                    UBI_Items[key] = nil;
                end;
            end;
        end;
    end;

-- Update inventory
    function UberInventory_UpdateInventory()
        -- Update the inventory
        UberInventory_Save_Bank();
        UberInventory_Save_Bag();
        UberInventory_Save_Keyring();
        UberInventory_Save_Mailbox();
        UberInventory_Save_Equipped();

        -- Cleanup the inventory (remove zero-count items)
        UberInventory_Cleanup();

        -- Get cash information
        UberInventory_GetMoney();

        -- Update wallet/mail info (if mainframe is shown)
        UberInventory_WalletMailCashInfo();

        -- Update inventory frame if visible
        if ( UberInventoryFrame:IsVisible() ) then
            -- Update items
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;

        -- Perform garbage collection
        collectgarbage( "collect" );
    end;

-- Handle UBI initialization
    function UberInventory_Init()
        -- Initiliaze all parts of the saved variable
        if (not UBI_Data) then
            UBI_Data = {};
        end;

        if (not UBI_Data["ItemPrices"]) then
            UBI_Data["ItemPrices"] = {};
        end;

        if (not UBI_Data[UBI_REALM]) then
            UBI_Data[UBI_REALM] = {};
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]) then
            UBI_Data[UBI_REALM][UBI_PLAYER] = {};
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]["Options"]) then
            UBI_Data[UBI_REALM][UBI_PLAYER]["Options"] = { ["version"] = UBI_VERSION,
                                                           ["show_money"] = true,
                                                           ["show_balance"] = true,
                                                           ["show_sell_prices"] = true,
                                                           ["show_minimap"] = true,
                                                           ["bag_max"] = 0,
                                                           ["bag_free"] = 0,
                                                           ["bank_max"] = 0,
                                                           ["bank_free"] = 0,
                                                           ["minimap"] = 315,
                                                           ["alpha"] = 1,
                                                           ["take_money"] = false,
                                                           ["faction"] = UBI_FACTION };
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]["Items"]) then
            UBI_Data[UBI_REALM][UBI_PLAYER]["Items"] = {};
        end;

        if (not UBI_Data[UBI_REALM][UBI_PLAYER]["Money"]) then
            UBI_Data[UBI_REALM][UBI_PLAYER]["Money"] = { ["mail"] = 0,
                                                         ["Cash"] = 0 };
        end;

        if (not UBI_Data[UBI_REALM]["Guildbank"]) then
            UBI_Data[UBI_REALM]["Guildbank"] = {};
        end;

        -- Create global variables for easy reference
        UBI_Money = UBI_Data[UBI_REALM][UBI_PLAYER]["Money"];
        UBI_Items = UBI_Data[UBI_REALM][UBI_PLAYER]["Items"];
        UBI_Options = UBI_Data[UBI_REALM][UBI_PLAYER]["Options"];
        UBI_Prices = UBI_Data["ItemPrices"];
        UBI_Guildbank = UBI_Data[UBI_REALM]["Guildbank"];

        -- Upgrade data structures
        UberInventory_Upgrade();

        -- Build list of alts and guildbanks
        UberInventory_GetAlts();
        UberInventory_GetGuildbanks();

        -- Build the inventory
        UberInventory_UpdateInventory();

        -- Initialize Options dialog
        UberInventoryOptionsFrame:Show();
    end;

-- Initialize guildbank data structure
    function UberInventory_Guildbank_Init()
        if ( IsInGuild() ) then
            if ( not UBI_Guildbank[UBI_GUILD] ) then
                UBI_Guildbank[UBI_GUILD] = {};
                UBI_Guildbank[UBI_GUILD]["Items"] = {};
                UBI_Guildbank[UBI_GUILD]["Cash"] = 0;
                UBI_Guildbank[UBI_GUILD]["Faction"] = UBI_FACTION;
            end;
        end;
    end;

-- Register events
    function UberInventory_RegisterEvents()
        foreach( UBI_TRACKED_EVENTS, function( key, value )
            this:RegisterEvent( value );
        end );
    end;

-- Unregister events
    function UberInventory_UnregisterEvents()
        foreach( UBI_TRACKED_EVENTS, function( key, value )
            this:UnregisterEvent( value );
        end );
    end;

-- Handle startup of the addon
    function UberInventory_OnLoad()
        -- Register slash commands
        SlashCmdList["UBI"] = UberInventory_SlashHandler;
        SLASH_UBI1 = "/ubi";
        SLASH_UBI2 = "/uberinventory";

        -- Register events to be monitored
        UberInventory_RegisterEvents();

       -- Correct quality list
       table.insert( UBI_QUALITY, 1, UBI_ALL_QUALITIES );
    end;

-- Install all of the hooks used by UberInventory
    function UberInventory_Install_Hooks()
        -- Save original GameTooltip function for OnTooltipAddMoney and set new handler
        UberInventory_ItemTooltip_Original = UberInventory_ItemTooltip:GetScript( "OnTooltipAddMoney" );
        UberInventory_ItemTooltip:SetScript( "OnTooltipAddMoney", UberInventory_OnTooltipAddMoney );

        -- Hook the Tooltips (OnTooltipSetItem, OnTooltipCleared)
        UberInventory_HookTooltip( GameTooltip );
        UberInventory_HookTooltip( ItemRefTooltip );
        UberInventory_HookTooltip( ShoppingTooltip1 );
        UberInventory_HookTooltip( ShoppingTooltip2 );
    end;

-- Handle events sent to the addon
    function UberInventory_OnEvent( self, event, ... )
        -- Track event count
        UBI_Events[event] = ( UBI_Events[event] or 0) + 1;

        -- When debug is on report event and event counts
        if ( UBI_DEBUG ) then
            -- Parse all arguments
            local argData = "";
            for ct = 1, select( '#', ... ) do
                local data = select( ct, ... )
                if type( data ) == "boolean" then
                    if ( data ) then
                        argData = argData.."["..ct.."] = <True> ";
                    else
                        argData = argData.."["..ct.."] = <False> ";
                    end;
                else
                    argData = argData.."["..ct.."] = "..data.." ";
                end;
            end;

            -- Print the data
            UberInventory_Message( event.." ("..UBI_Events[event]..") = "..argData );
        end;

        -- Show message that this AddOn has been loaded
        if ( event == "ADDON_LOADED" and arg1 == "UberInventory" ) then
            -- Show startup message
            UberInventory_Message( UBI_STARTUP_MESSAGE, false );

            -- Install hooks
            UberInventory_Install_Hooks();
            return;
        end;

        -- Load data from previous session
        if ( event == "VARIABLES_LOADED" ) then
            UberInventory_Init();

            -- Check bank visit
            if ( UberInventory_DaysSince( UBI_Options["bank_visit"] ) >= UBI_BANK_VISIT_INTERVAL ) then
                UberInventory_Message( UBI_LAST_BANK_VISIT:format( UBI_Options["bank_visit"] ) );
            else
                if ( not UBI_Options["bank_visit"] ) then UberInventory_Message( UBI_VISIT_BANK ); end;
            end;

            -- Check guildbank visit
            if ( IsInGuild() ) then
                if ( UberInventory_DaysSince( UBI_Options["guildbank_visit"] ) >= UBI_GUILDBANK_VISIT_INTERVAL ) then
                    UberInventory_Message( UBI_LAST_GUILDBANK_VISIT:format( UBI_Options["guildbank_visit"] ) );
                else
                    if ( not UBI_Options["guildbank_visit"] ) then UberInventory_Message( UBI_VISIT_GUILDBANK ); end;
                end;
            end;

            -- Check mailbox visit
            if ( UberInventory_DaysSince( UBI_Options["mail_visit"] ) >= UBI_MAILBOX_VISIT_INTERVAL ) then
                UberInventory_Message( UBI_LAST_MAILBOX_VISIT:format( UBI_Options["mail_visit"] ) );
            else
                if ( not UBI_Options["mail_visit"] ) then UberInventory_Message( UBI_VISIT_MAILBOX ); end;
            end;

            -- Show or hide the minimap icon
            UberInventory_Minimap_Toggle();

            -- Assign proper location for minimap
            UberInventory_Minimap_Update( UBI_Options["minimap"] );

            -- Set transparency
            UberInventory_Change_Alpha( UBI_Options["alpha"] );

            -- Broadcast current version to the guild
            UberInventory_SendVersion();
            return;
        end;

        -- Update info inventory change (enters the world, enters/leaves an instance, or respawns at a graveyard, reloadui)
        if ( event == "PLAYER_ENTERING_WORLD" or ( event == "UNIT_INVENTORY_CHANGED" and arg1 == "player" ) ) then
            -- Update balance frame
            UberInventory_ShowBalanceFrame( UBI_Options["show_balance"] );

            -- Get guild name
            UBI_GUILD = GetGuildInfo( "player" );

            UberInventory_UpdateInventory();
            return;
        end;

        -- After a money event show the current cash amount
        if ( event == "PLAYER_MONEY" ) then
            -- Save current cash balance
            UberInventory_GetMoney();

            -- Update wallet/mail info (if mainframe is shown)
            UberInventory_WalletMailCashInfo();

            -- Show message
            if ( UBI_Options["show_money"] ) then
                UberInventory_Message( UBI_MONEY_MESSAGE:format( UberInventory_GetMoneyString( GetMoney() ) ) );
            end;
            return;
        end;

        -- Bank closed
        if ( event == "BANKFRAME_CLOSED" ) then
            UBI_BANK_OPEN = false;
            return;
        end;

        -- Bank opened
        if ( event == "BANKFRAME_OPENED" ) then
            UBI_BANK_OPEN = true;
            UBI_Options["bank_visit"] = time();
            UberInventory_UpdateInventory();
            return;
        end;

        -- Items in the bank have changed
        if ( event == "PLAYERBANKSLOTS_CHANGED" or event == "PLAYERBANKBAGSLOTS_CHANGED" ) then
            UberInventory_UpdateInventory();
            return;
        end;

        -- Guildbank closed (send guildbank data using the chat system)
        if ( event == "GUILDBANKFRAME_CLOSED" ) then
            -- Store guildbank cash
            UBI_Guildbank[UBI_GUILD]["Cash"] = GetGuildBankMoney();

            -- Trick to work around the issue that the GUILDBANKFRAME_CLOSED event is triggered twice when closing the Guildbank
            if ( not UBI_GUILDBANK_OPEN ) then UberInventory_Send_Guildbank(); end;
            UBI_GUILDBANK_OPEN = false;
            return;
        end;

        -- Guild cash changed (When does this event occur?)
        if ( event == "GUILDBANK_UPDATE_MONEY" and IsInGuild() ) then
            UBI_Guildbank[UBI_GUILD]["Cash"] = GetGuildBankMoney();

            -- Update cash info on screen
            if ( UberInventoryFrame:IsVisible() ) then
                UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
            end;
        end;

        -- Guildbank opened
        if ( event == "GUILDBANKFRAME_OPENED" ) then
            -- Track Guildbank open
            UBI_GUILDBANK_OPEN = true;

            -- Store guildbank cash
            UBI_Guildbank[UBI_GUILD]["Cash"] = GetGuildBankMoney();

            -- Track Guildbank last visit date/time
            UBI_Options["guildbank_visit"] = time();

            -- Store Guildbank data
            UberInventory_Save_Guildbank( event );
            return;
        end;

        -- Update Guildbank information
        if ( ( event == "GUILDBANK_UPDATE_TABS" or event == "GUILDBANKBAGSLOTS_CHANGED" ) and UBI_GUILDBANK_OPEN ) then
            UberInventory_Save_Guildbank( event );
            return;
        end;

        -- Mailbox opened
        if ( event == "MAIL_SHOW" ) then
            UBI_MAILBOX_OPEN = true;
            UBI_Options["mail_visit"] = time();
            UberInventory_UpdateInventory();
            return;
        end;

        -- Mailbox changes
        if ( event == "UPDATE_PENDING_MAIL" or event == "MAIL_INBOX_UPDATE" ) then
            UberInventory_UpdateInventory();
            return;
        end;

        -- Mailbox closed
        if ( event == "MAIL_CLOSED" ) then
            UBI_MAILBOX_OPEN = false;
            return;
        end;

        -- Merchant window opened
        if ( event == "MERCHANT_SHOW" ) then
            -- Update merchant buy/sell prices
            UberInventory_MerchantUpdate();
            return;
        end;

        -- Items in the bags have changed
        if ( event == "BAG_UPDATE" or event == "ITEM_PUSH" ) then
            -- Update inventory
            UberInventory_UpdateInventory();

            -- Update merchant sell/buy prices
            if ( MerchantFrame:IsVisible() ) then
                UberInventory_MerchantUpdate();
            end;
            return;
        end;

        -- Guild changes
        if ( event == "PLAYER_GUILD_UPDATE" and IsInGuild() ) then
            -- Get guild name
            UBI_GUILD = GetGuildInfo( "player" );

            -- Rebuild list of guildbanks
            UberInventory_GetGuildbanks();

            UberInventory_Locations_Init();
            return;
        end;

        -- Handle guild roster changes
        if ( event == "GUILD_ROSTER_UPDATE" and IsInGuild()  ) then
            -- Get guild name
            UBI_GUILD = GetGuildInfo( "player" );

            -- Set faction for guildbank
            if ( UBI_GUILD ) then
                -- Initialize guildbank (if needed)
                UberInventory_Guildbank_Init();

                UBI_Guildbank[UBI_GUILD]["Faction"] = UBI_FACTION;
                UberInventory_GetGuildbanks();
            end;
        end;

        -- Handle UberInventory related guildchat messages
        if ( event == "CHAT_MSG_ADDON" and arg1 == "UBI" and arg4 ~= UBI_PLAYER ) then
            if ( arg2:find( "UBI:VERSION" ) ) then
                -- Extract version number from the data string
                local _, gbVersion = strsplit( " ", arg2 );

                -- Check if there is a more recent version of UberInventory
                if ( tonumber( gbVersion ) > ( tonumber( UBI_VERSION ) or 0 ) and not UBI_VERSION_WARNING ) then
                    UberInventory_Message( UBI_NEW_VERSION:format( arg4, gbVersion ) );
                    UBI_VERSION_WARNING = true;
                end;

                -- If this version is more recent send message to other users using this addon
                if ( tonumber( gbVersion ) < ( tonumber( UBI_VERSION ) or 0 ) ) then
                    UberInventory_SendVersion();
                end;
            elseif ( arg2:find( "UBI:GBSTART" ) and not UBI_GUILDBANK_OPEN ) then
                -- Start receiving Guildbank data, reset on each new receive from player.
                -- This way only the most recent data should be received.
                -- Only data from equal versions are allowed
                UBI_GBData = {};
                UBI_GBSender = arg4;
            elseif ( arg2:find( "UBI:GBITEM" ) and not UBI_GUILDBANK_OPEN ) then
                -- Receiving Guildbank data (item), only receive data from latest sender
                if ( UBI_GBSender == arg4 ) then
                    local _, itemID, itemCount, itemName, itemType, itemSubtype, itemLevel, itemQuality = strsplit( " ", arg2 );
                    UBI_GBData[itemID] = { ["itemid"] = itemID,
                                           ["count"] = tonumber( itemCount ),
                                           ["name"] = string.gsub( itemName, "_", " " ),
                                           ["type"] = string.gsub( itemType, "_", " " ),
                                           ["subtype"] = string.gsub( itemSubtype, "_", " " ),
                                           ["level"] = tonumber( itemLevel ),
                                           ["quality"] = tonumber( itemQuality ) };
                end;
            elseif ( arg2:find( "UBI:GBCASH" ) and not UBI_GUILDBANK_OPEN ) then
                -- Receiving Guildbank cash, only receive data from latest sender
                if ( UBI_GBSender == arg4 ) then
                    -- Init guildbank structures (if needed)
                    UberInventory_Guildbank_Init()

                    local _, gbCash = strsplit( " ", arg2 );
                    UBI_Guildbank[UBI_GUILD]["Cash"] = tonumber( gbCash );

                    -- Update cash info on screen
                    if ( UberInventoryFrame:IsVisible() ) then
                        UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT );
                    end;
                end;
            elseif ( arg2:find( "UBI:GBEND" ) and not UBI_GUILDBANK_OPEN ) then
                -- Stop receiving Guildbank data and process the received data, only process data from latest sender
                if ( UBI_GBSender == arg4 ) then
                    UberInventory_Receive_Guildbank();
                end;
            end;
        end;
    end;

-- Show or hide the balance frame
    function UberInventory_ShowBalanceFrame( showFrame )
        if ( showFrame ) then
            UberInventoryBalanceFrame:Show();
            MoneyFrame_Update( "UberInventoryBalanceFrameMoneyFrame", GetMoney() );
        else
            UberInventoryBalanceFrame:Hide();
        end;
    end;

-- Update minimap button
    function UberInventory_Minimap_Update( angle )
        UberInventoryMinimap:SetPoint( "TOPLEFT",
                                       "Minimap",
                                       "TOPLEFT",
                                       54 - (78 * cos( angle )),
                                       (78 * sin( angle )) - 55 );
    end;

-- Update minimap button
    function UberInventory_Minimap_Toggle()
        if ( UBI_Options["show_minimap"]) then
            UberInventoryMinimap:Show();
        else
            UberInventoryMinimap:Hide();
        end;
    end;

-- Update transparency of frames
    function UberInventory_Change_Alpha( alpha )
        UberInventoryFrameTexture:SetAlpha( alpha );
        UberInventoryFrameBorderTexture:SetAlpha( alpha );
    end;

-- Init frame options
    function UberInventory_SetOptions( frame )
        getglobal( frame:GetName().."Money" ):SetChecked( UBI_Options["show_money"] );
        getglobal( frame:GetName().."Balance" ):SetChecked( UBI_Options["show_balance"] );
        getglobal( frame:GetName().."SellPrice" ):SetChecked( UBI_Options["show_sell_prices"] );
        getglobal( frame:GetName().."InboxMoney" ):SetChecked( UBI_Options["take_money"] );
        getglobal( frame:GetName().."ShowMinimap" ):SetChecked( UBI_Options["show_minimap"] );
        getglobal( frame:GetName().."Minimap" ):SetValue( UBI_Options["minimap"] );
        getglobal( frame:GetName().."Alpha" ):SetValue( UBI_Options["alpha"] );
    end;

-- Save frame options
    function UberInventory_SaveOptions( frame )
        UBI_Options["show_money"] = getglobal( frame:GetName().."Money" ):GetChecked();
        UBI_Options["show_balance"] = getglobal( frame:GetName().."Balance" ):GetChecked();
        UBI_Options["show_sell_prices"] = getglobal( frame:GetName().."SellPrice" ):GetChecked();
        UBI_Options["take_money"] = getglobal( frame:GetName().."InboxMoney" ):GetChecked();
        UBI_Options["show_minimap"] = getglobal( frame:GetName().."ShowMinimap" ):GetChecked();
        UBI_Options["minimap"] = getglobal( frame:GetName().."Minimap" ):GetValue();
        UBI_Options["alpha"] = getglobal( frame:GetName().."Alpha" ):GetValue();
    end;

-- Apply setting to the UI
    function UberInventory_UpdateUI()
        -- Hide or show balance frame based on the current settings (look for a more elegant solution)
        UberInventory_ShowBalanceFrame( UBI_Options["show_balance"] );

        -- Show or hide minimap icon
        UberInventory_Minimap_Toggle();

        -- Restore or set minimap location
        UberInventory_Minimap_Update( UBI_Options["minimap"] );

        -- Restore or set transparency
        UberInventory_Change_Alpha( UBI_Options["alpha"] );
    end;

-- Revert all options to default values
    function UberInventory_SetDefaults( frame )
        -- Restore defaults
        UBI_Options["show_money"] = true;
        UBI_Options["show_balance"] = true;
        UBI_Options["show_sell_prices"] = true;
        UBI_Options["show_minimap"] = true;
        UBI_Options["minimap"] = 315;
        UBI_Options["alpha"] = 1;
        UBI_Options["take_money"] = false;

        -- Hide or show balance frame based on the current settings (look for a more elegant solution)
        UberInventory_ShowBalanceFrame( UBI_Options["show_balance"] );

        -- Show or hide minimap icon
        UberInventory_Minimap_Toggle();

        -- Restore or set minimap location
        UberInventory_Minimap_Update( UBI_Options["minimap"] );

        -- Restore or set transparency
        UberInventory_Change_Alpha( UBI_Options["alpha"] );

        -- Redraw Options frame
        UberInventory_SetOptions( frame );
    end;

-- Make Settings frame known to the system
    function UberInventory_AddCategory( frame )
        -- Set panel name
        frame.name = UBI_NAME;

        -- Set okay function
        frame.okay = function( self )
            UberInventory_SaveOptions( self );
            UberInventory_UpdateUI();
        end;

        -- Set cancel function
        frame.cancel = function( self )
            UberInventory_SetOptions( self );
            UberInventory_UpdateUI();
        end;

        -- set defaults function
        frame.default = function( self )
            UberInventory_SetDefaults( self );
            UberInventory_UpdateUI();
        end;

        -- Add the panel
        InterfaceOptions_AddCategory( frame );
    end;

-- Update inventory button
    function UberInventory_UpdateButton( slot, record )
        local sellPrice, buyPrice, totalCount = 0, 0, 0;
        local objSlot = getglobal( "UberInventoryFrameItem"..slot );
        local button = objSlot:GetName().."ItemButton";
        local buttonObj = getglobal( button );

        -- Record is nil, item does not exist and button should be hidden
        if ( not record ) then
            objSlot:Hide();
            return;
        else
            objSlot:Show();
        end;

        -- Get itemID
        local itemid = record["itemid"];

        -- Get item prices
        if ( UBI_Prices[itemid] ) then
            if ( UBI_Prices[itemid]["sell"] ) then sellPrice = UBI_Prices[itemid]["sell"]; end;
            if ( UBI_Prices[itemid]["buy"] ) then buyPrice = UBI_Prices[itemid]["buy"]; end;
        end;

        -- Get item counts
        local bagCount = record["bag_count"] or 0;
        local bankCount = record["bank_count"] or 0;
        local keyCount = record["keyring_count"] or 0;
        local mailCount = record["mailbox_count"] or 0;
        local equipCount = record["equip_count"] or 0;
        local guildCount = record["count"] or 0;
        local totalCount = record["total"] or 0;

        -- Set item image
        SetItemButtonTexture( buttonObj, UberInventory_GetTexture( itemid ) );

        -- Set item name
        getglobal( button.."ItemName" ):SetText( record["name"] );
        getglobal( button.."ItemName" ):SetTextColor( ITEM_QUALITY_COLORS[record.quality].r,
                                                      ITEM_QUALITY_COLORS[record.quality].g,
                                                      ITEM_QUALITY_COLORS[record.quality].b );

        -- Set item count
        if ( UBI_SHOWING_GUILDBANK ) then
            getglobal( button.."ItemCount" ):SetFormattedText( UBI_ITEM_COUNT_SINGLE, totalCount );
        else
            getglobal( button.."ItemCount" ):SetFormattedText( UBI_ITEM_COUNT, totalCount, bagCount, bankCount, keyCount, mailCount, equipCount );
        end;
        SetItemButtonCount( buttonObj, totalCount );

        -- Set item id (for tooltip)
        buttonObj:SetID( itemid );
        buttonObj:SetAttribute( "inventoryitem", record );
        buttonObj:SetAttribute( "location", UBI_FILTER_LOCATIONS );

        -- Set item buy price
        if ( buyPrice > 0 ) then
          getglobal( button.."ItemBuy" ):SetText( UBI_ITEM_BUY..UberInventory_GetMoneyString( buyPrice ) );
        else
          getglobal( button.."ItemBuy" ):SetText( UBI_ITEM_BUY.."-" );
        end;

        -- Set item sell price
        if ( sellPrice > 0 ) then
          getglobal( button.."ItemSell" ):SetText( UBI_ITEM_SELL..UberInventory_GetMoneyString( sellPrice ) );
        else
          getglobal( button.."ItemSell" ):SetText( UBI_ITEM_SELL.."-" );
        end;

        -- Mark usability onto the itembutton
        if UberInventory_UsableItem( itemid ) then
            SetItemButtonTextureVertexColor( buttonObj, 1, 1, 1 );
        else
            SetItemButtonTextureVertexColor( buttonObj, 1, 0.1, 0.1 );
        end;
    end;

-- Display items (by page, UBI_NUM_ITEMBUTTONS items)
    function UberInventory_DisplayItems( page, str )
        -- Create GB structure if it does not already exist
        UberInventory_Guildbank_Init();

        -- Sort and filter items
        UberInventory_SortFilter( str or "", false );

        -- Update scrollbars
        if ( UBI_DISP_COUNT ~= #UBI_Sorted ) then
            UberInventory_Scroll_Init( math.ceil( #UBI_Sorted / UBI_NUM_ITEMBUTTONS ) - 1 );
        end;
        UBI_DISP_COUNT = #UBI_Sorted;

        -- Calculate start item
        local startItem = page * UBI_NUM_ITEMBUTTONS + 1;

        -- Execute code to actually add the items to the inventory frame
        for slot = 1, UBI_NUM_ITEMBUTTONS do
            -- Calculate the actual index within sorted array
            local index = startItem + slot - 1;

            -- Display item
            UberInventory_UpdateButton( slot, UBI_Sorted[ index ] );
        end;

        -- Update inventory count
        UberInventoryFrameInventoryCount:SetFormattedText( UBI_INVENTORY_COUNT, #UBI_Sorted, UBI_Inventory_count );

        -- Show/Hide the moneyframe for displaying Guildbank or Alt cash
        if ( UBI_FILTER_LOCATIONS < 200 ) then
            UberInventoryFrameMoneyOthersCash:SetText( UBI_EMPTY_TEXT );
        else
            -- Initialize
            local otherCash = 0;

            -- Determine the cash owned
            if ( UBI_FILTER_LOCATIONS >= 1000 ) then
               otherCash = UBI_Guildbank[UBI_Guildbanks[UBI_FILTER_LOCATIONS - 999]]["Cash"];
            elseif ( UBI_FILTER_LOCATIONS >= 200 ) then
               otherCash = UBI_Data[UBI_REALM][UBI_Characters[UBI_FILTER_LOCATIONS - 199]]["Money"]["Cash"];
            end;

            -- Update the cash info for Alt/Guildbank
            UberInventoryFrameMoneyOthersCash:SetText( UberInventory_GetMoneyString( otherCash ) );
        end;
    end;

-- Update wallet/mail cash info on screen
    function UberInventory_WalletMailCashInfo()
        if(  UberInventoryFrame:IsVisible() ) then
            -- Update cash balance information
            UberInventoryFrameMoneyWalletCash:SetText( UberInventory_GetMoneyString( GetMoney() ) );
            UberInventoryFrameMoneyMailCash:SetText( UberInventory_GetMoneyString( UBI_Money["mail"] ) );
        end;
    end;

-- Show or hide the Main dialog
    function UberInventory_Toggle_Main()
        if(  UberInventoryFrame:IsVisible() ) then
            UberInventoryFrame:Hide();
        else
            -- Show the frame
            UberInventoryFrame:Show();
            UberInventoryFrame:SetBackdropColor( 0, 0, 0, 1 );

            -- Update wallet/mail cash information
            UberInventory_WalletMailCashInfo();

            -- Force refresh
            UberInventory_DisplayItems( UBI_PAGE, UBI_FILTER_TEXT or "" );

            -- Update slot count information
            UberInventory_Update_SlotCount();
        end;
    end;

-- Perform item search
    function UberInventory_Search( str )
        -- Display search criteria
        UberInventory_Message( UBI_ITEM_SEARCH:format( str ) );

        -- Sort and filter items
        UberInventory_SortFilter( str or "", true );

        -- Search and display items found
        foreach( UBI_Sorted, function( key, value )
            -- Get item count
            local itemCounts = { value["total"],
                                 value["bag_count"] or 0,
                                 value["bank_count"] or 0,
                                 value["keyring_count"] or 0,
                                 value["mailbox_count"] or 0,
                                 value["equip_count"] or 0 };

            -- Build search result
            local itemText = "|T"..UberInventory_GetTexture( value["itemid"] )..":0|t "..UberInventory_GetLink( value["itemid"] );
            local msg = "  "..itemText.." "..UBI_ITEM_COUNT_SINGLE:format( itemCounts[1] ).." (";
            for loc = 2, #itemCounts do
                if ( itemCounts[loc] > 0 ) then
                    msg = msg.."|T"..UBI_LOCATION_TEXTURE[loc]..":0|t "..UBI_LOCATIONS[loc].." = "..itemCounts[loc].." / ";
                end;
            end;

            -- Display result
            UberInventory_Message( msg:sub( 1, msg:len()-3 )..")" );
        end );

        -- Display end of search
        if ( #UBI_Sorted == 0 ) then
           UberInventory_Message( UBI_ITEM_SEARCH_NONE );
        else
           UberInventory_Message( UBI_ITEM_SEARCH_DONE );
        end;
    end;

-- Remove character or guildbank data
    function UberInventory_RemoveData( data )
        -- split of command and name (guildbank names can contain spaces, unlike character names)
        local command, _ = strsplit( " ", data ):lower();
        local name = data:sub( command:len()+2 );

        -- Remove the data
        if ( command == "remchar"  and name ~= "Guildbank" ) then
            -- Do not remove data for the current connected character
            if ( name == UBI_PLAYER ) then
                UberInventory_Message( UBI_REM_NOTALLOWED );
                return;
            end;

            -- Check if character has data stored and delete it if so
            if ( UBI_Data[UBI_REALM][name] ) then
                StaticPopupDialogs["UBI_CONFIRM_DELETE"].OnAccept = function()
                    -- Remove data
                    UBI_Data[UBI_REALM][name] = nil;

                    -- Display message
                    UberInventory_Message( UBI_REM_DONE:format( name ) );

                    -- Rebuild list of alts
                    UberInventory_GetAlts();

                    -- Reset filters
                    UberInventory_ResetFilters();
                end;

                -- Ask for remove confirmation
                StaticPopup_Show( "UBI_CONFIRM_DELETE", UBI_REM_CHARACTER:format( name ) );
            else
                UberInventory_Message( UBI_REM_CHARNOTFOUND:format( name ) );
            end;
        elseif ( command == "remguild" ) then
            -- Check if guildbank has data stored and delete it if so
            if ( UBI_Guildbank[name] ) then
                StaticPopupDialogs["UBI_CONFIRM_DELETE"].OnAccept = function()
                    -- Remove data
                    UBI_Guildbank[name] = nil;

                    -- Display message
                    UberInventory_Message( UBI_REM_DONE:format( name ) );

                    -- Rebuild list of guildbanks
                    UberInventory_GetGuildbanks();

                    -- Reset filters
                    UberInventory_ResetFilters();
                end;

                -- Ask for remove confirmation
                StaticPopup_Show( "UBI_CONFIRM_DELETE", UBI_REM_GUILDBANK:format( name ) );
            else
                UberInventory_Message( UBI_REM_GUILDNOTFOUND:format( name ) );
            end;
        end;
    end;

-- Show help message
    function UberInventory_ShowHelp()
        foreach( UBI_HELP, function( key, value )
            UberInventory_Message( value )
        end );
    end;

-- Show event counts
    function UberInventory_EventCounts()
        foreach( UBI_Events, function( key, value )
            UberInventory_Message( key.." = "..value );
        end );
    end;

-- Handle slash commands
    function UberInventory_SlashHandler( msg )
        -- arguments should be handled case-insensitve
        local orgData = msg;
        msg = string.lower( msg );

        -- Handle each individual argument
        if ( msg == "" ) then
            -- Show main dialog
            UberInventory_Toggle_Main();
        elseif ( msg:find( "remchar" ) or msg:find( "remguild" ) ) then
            -- Remove character or guild data
            UberInventory_RemoveData( orgData );
        elseif ( msg == "help" ) or ( msg == "?" ) then
            -- Show help info
            UberInventory_ShowHelp();
        elseif ( msg == "dump" ) then
            -- Show event counts
            UberInventory_EventCounts();
        elseif ( msg == "forceupdate" ) then
            -- Update stored inventory data
            UberInventory_UpdateInventory();
        elseif ( msg == "debug" ) then
            -- Toggle debug flag
            UBI_DEBUG = not UBI_DEBUG;
            if ( UBI_DEBUG ) then
                UberInventory_Message( "Debugging on" );
            else
                UberInventory_Message( "Debugging off" );
            end;
        else
            -- Perform item search
            UberInventory_Search( msg or "" );
        end;
    end;
